import numpy as np

from .base import io, tokenizers
from .base.common import load_numpy_array, map_checkbox_value, to_float, to_int
from .base.embeddings_model import (TFIDF, EmbeddingsModelBert,
                                    EmbeddingsModelBertEmbeddia,
                                    EmbeddingsModelDoc2Vec,
                                    EmbeddingsModelElmo,
                                    EmbeddingsModelFastText,
                                    EmbeddingsModelFastTextEmbedia,
                                    EmbeddingsModelFastTextHr,
                                    EmbeddingsModelFastTextSl,
                                    EmbeddingsModelGloVe, EmbeddingsModelLSI,
                                    EmbeddingsModelUniversalSentenceEncoder,
                                    EmbeddingsModelWord2Vec)
from .base.table import orange_data_table


def cf_text_embeddings_parse_csv(input_dict):
    filename = input_dict['input']
    text_index = input_dict.get('text_index')
    label_index = input_dict.get('label_index')
    delimiter = input_dict.get('delimiter') or None
    skip_header = input_dict.get('skip_header')
    skip_header = map_checkbox_value(skip_header)

    texts, labels = io.read_csv(filename=filename, delimiter=delimiter, skip_header=skip_header,
                                text_index=text_index, label_index=label_index)
    return {'texts': texts, 'labels': labels}


def cf_text_embeddings_tok_tok_tokenizer(input_dict):
    assert 'texts' in input_dict, 'Text is missing'
    assert 'to_lowercase' in input_dict, 'To lowercase argument is missing'

    to_lowercase = map_checkbox_value(input_dict['to_lowercase'])
    texts = tokenizers.toktok_tokenizer(input_dict['texts'], to_lowercase=to_lowercase)
    return {'texts': texts}


def cf_text_embeddings_punkt_tokenizer(input_dict):
    assert 'texts' in input_dict, 'Text is missing'

    language = input_dict['language']
    input_dict['texts'] = tokenizers.punkt_tokenizer(input_dict['texts'], language)
    return input_dict


def cf_text_embeddings_token_filtering(input_dict):
    assert 'texts' in input_dict, 'Text is missing'

    input_dict['texts'] = tokenizers.token_filtering(input_dict['texts'],
                                                     filter_tokens=input_dict['filter_tokens'])
    return input_dict


def cf_text_embeddings_tfidf(input_dict):
    assert 'texts' in input_dict, 'Text is missing'

    tfidf = TFIDF()
    tfidf.train(input_dict['texts'])
    return {'tfidf': tfidf}


def cf_text_embeddings_base(klass, input_dict):
    lang = input_dict.get('lang_selector') or input_dict.get('lang')
    embeddings_model = klass(lang)

    aggregation_method = input_dict.get('aggregation_method')
    tfidf = input_dict.get('tfidf')
    texts = input_dict['texts']
    embeddings = embeddings_model.apply(texts, aggregation_method, tfidf=tfidf)
    return {'embeddings': embeddings}


def cf_text_embeddings_word2vec(input_dict):
    return cf_text_embeddings_base(EmbeddingsModelWord2Vec, input_dict)


def cf_text_embeddings_glove(input_dict):
    return cf_text_embeddings_base(EmbeddingsModelGloVe, input_dict)


def cf_text_embeddings_fasttext(input_dict):
    return cf_text_embeddings_base(EmbeddingsModelFastText, input_dict)


def cf_text_embeddings_fasttext_sl(input_dict):
    input_dict['lang'] = 'sl'
    return cf_text_embeddings_base(EmbeddingsModelFastTextSl, input_dict)


def cf_text_embeddings_fasttext_embeddia(input_dict):
    return cf_text_embeddings_base(EmbeddingsModelFastTextEmbedia, input_dict)


def cf_text_embeddings_fasttext_hr(input_dict):
    input_dict['lang'] = 'hr'
    return cf_text_embeddings_base(EmbeddingsModelFastTextHr, input_dict)


def cf_text_embeddings_elmo(input_dict):
    return cf_text_embeddings_base(EmbeddingsModelElmo, input_dict)


def cf_text_embeddings_lsi(input_dict):
    texts = input_dict['texts']
    num_topics_default = 200
    num_topics = to_int(input_dict['num_topics'], num_topics_default)
    num_topics = num_topics_default if num_topics < 1 else num_topics
    decay = to_float(input_dict['decay'], 1.0)
    train_on_tfidf = map_checkbox_value(input_dict['train_on_tfidf'])
    filter_extremes = map_checkbox_value(input_dict['filter_extremes'])

    embeddings_model = EmbeddingsModelLSI(num_topics, decay, train_on_tfidf,
                                          filter_extremes=filter_extremes)
    embeddings = embeddings_model.apply(texts)
    return {'embeddings': embeddings}


def cf_text_embeddings_bert(input_dict):
    texts = input_dict['texts']

    model = input_dict['model_selection']
    embeddings_model = EmbeddingsModelBert(model)
    embeddings = embeddings_model.apply(texts)
    return {'embeddings': embeddings}


def cf_text_embeddings_bert_embeddia(input_dict):
    texts = input_dict['texts']

    model = input_dict['model_selection']
    embeddings_model = EmbeddingsModelBertEmbeddia(model)
    embeddings = embeddings_model.apply(texts)
    return {'embeddings': embeddings}


def cf_text_embeddings_universal_sentence_encoder(input_dict):
    return cf_text_embeddings_base(EmbeddingsModelUniversalSentenceEncoder, input_dict)


def cf_text_embeddings_doc2vec(input_dict):
    lang = input_dict.get('lang_selector') or input_dict.get('lang')
    texts = input_dict['texts']

    embeddings_model = EmbeddingsModelDoc2Vec(lang)
    embeddings = embeddings_model.apply(texts)
    return {'embeddings': embeddings}


def cf_text_embeddings_language(input_dict):
    return input_dict


def cf_text_embeddings_import_dataset(input_dict):
    x_path = input_dict['x_path']
    y_path = input_dict['y_path']

    embeddings = load_numpy_array(x_path)
    labels = load_numpy_array(y_path)
    dataset = orange_data_table(embeddings, labels)
    return {'dataset': dataset}


def cf_text_embeddings_create_dataset(input_dict):
    embeddings = input_dict['embeddings']
    labels = input_dict['labels']
    dataset = orange_data_table(embeddings, labels)
    return {'dataset': dataset}


def cf_text_embeddings_export_dataset(input_dict):
    return input_dict


def cf_text_embeddings_concatenate_embeddings(input_dict):
    arrays = input_dict['embedding']
    return {'concat_embeddings': np.concatenate(arrays, axis=1)}


def cf_text_embeddings_make_scikit_bunch(input_dict):
    from sklearn.datasets import base as ds
    dataset = ds.Bunch(data=input_dict['X'], target=input_dict['y'],
                       feature_names=input_dict['feature_names'], DESCR=input_dict['description'])
    return {'dataset': dataset}



import unittest

from ..base.io import parse_csv, validate_and_convert_indices


class TestParseCSV(unittest.TestCase):
    def test_parse_csv(self):
        rows = [['text', 'label'], ['This is some text', '1'], ['This is some other text', '0']]
        texts, labels = parse_csv(rows, 1, 2)
        dc = {'text': texts, 'label': labels}

        for i, column in enumerate(['text', 'label']):
            self.assertEqual(len(rows), len(dc[column]))
        for i, row in enumerate(rows):
            for j, column in enumerate(['text', 'label']):
                self.assertEqual(row[j], dc[column][i])

    def test_multiple_spaces_in_text(self):
        rows = [['text', 'label'], ['This is    some text ', '1'],
                ['This is some      other text', '0']]
        texts, labels = parse_csv(rows, 1, 2)
        dc = {'text': texts, 'label': labels}

        for column in ['text', 'label']:
            self.assertEqual(len(rows), len(dc[column]))
        for i, row in enumerate(rows):
            for j, column in enumerate(['text', 'label']):
                self.assertEqual(row[j], dc[column][i])

    def test_no_label(self):
        rows = [['text'], ['This is some text'], ['This is some other text']]
        texts, labels = parse_csv(rows, 1, None)
        dc = {'text': texts, 'label': labels}

        self.assertEqual(dc['label'], [])
        for i, column in enumerate(['text']):
            self.assertEqual(len(rows), len(dc[column]))
        for i, row in enumerate(rows):
            for j, column in enumerate(['text']):
                self.assertEqual(row[j], dc[column][i])


class ValidateAndConvertIndices(unittest.TestCase):
    def test_all_indices(self):
        text_index, label_index = validate_and_convert_indices(1, 2)
        self.assertEqual(text_index, 1)
        self.assertEqual(label_index, 2)

    def test_all_indices_string(self):
        text_index, label_index = validate_and_convert_indices('1', '2')
        self.assertEqual(text_index, 1)
        self.assertEqual(label_index, 2)

    def test_zero_index(self):
        self.assertRaises(Exception, validate_and_convert_indices, '0', '1')

    def test_label_empty_string(self):
        text_index, label_index = validate_and_convert_indices('1', '')
        self.assertEqual(text_index, 1)
        self.assertEqual(label_index, None)

    def test_label_letters(self):
        text_index, label_index = validate_and_convert_indices('1', 'b')
        self.assertEqual(text_index, 1)
        self.assertEqual(label_index, None)

    def test_no_text(self):
        self.assertRaises(Exception, validate_and_convert_indices, '', '')


if __name__ == '__main__':
    unittest.main()



from nltk.data import load
from nltk.tokenize import ToktokTokenizer


def toktok_tokenizer(documents, to_lowercase=True):
    tokenizer = ToktokTokenizer()
    documents_tokens = []
    for document in documents:
        if to_lowercase:
            document = document.lower()
        document_tokens = tokenizer.tokenize(document)
        documents_tokens.append(document_tokens)
    return documents_tokens


def punkt_tokenizer(documents, language):
    supported_languages = {'english', 'estonian', 'german', 'slovene', 'spanish'}
    assert language in supported_languages, ('Language %s is not supported by punkt tokenizer' %
                                             language)

    tokenizer = load('tokenizers/punkt/{0}.pickle'.format(language))

    documents_tokens = []
    for document in documents:
        document_tokens = tokenizer.tokenize(document)
        documents_tokens.append(document_tokens)
    return documents_tokens


def token_filtering(documents_tokens, filter_tokens=None):
    default_filter = {
        '(', ')', '[', ']', '{', '}', "'", '"', ';', ':', ',', '!', '/', '\\', '@', '#', '$', '^',
        '*', '+', '=', '-', '|', '_', '<', '>', '?', '!', '.', ',', '&', '%', '0', '1', '2', '3',
        '4', '5', '6', '7', '8', '9'
    }
    filter_ = default_filter if not filter_tokens else set(
        [tok.strip() for tok in filter_tokens.split('\n')])

    documents_tokens_filtered = []
    for document_tokens in documents_tokens:
        document_tokens_filtered = [
            document_token for document_token in document_tokens if document_token not in filter_
        ]
        documents_tokens_filtered.append(document_tokens_filtered)
    return documents_tokens_filtered



from Orange.data import ContinuousVariable, DiscreteVariable, Domain, Table


# def orange_domain(n_features, unique_labels):
#     return Domain([ContinuousVariable.make('Feature %d' % i) for i in range(n_features)],
#                   DiscreteVariable('class', values=unique_labels))


def orange_data_table(embeddings, labels):
    if labels == [] or labels is None:
        domain = Domain([ContinuousVariable.make('Feature %d' % i) for i in range(embeddings.shape[1])])
        y = None
    else:
        uniq_labels = list(sorted(set(labels)))
        y = [uniq_labels.index(r) for r in labels]
        domain = Domain([ContinuousVariable.make('Feature %d' % i) for i in range(embeddings.shape[1])],
                        DiscreteVariable('class', values=uniq_labels))
    table = Table(domain, embeddings, Y=y)
    return table



import csv


def map_delimiter(delimiter):
    if delimiter == '\\t':
        return '\t'
    return delimiter


def read_file(filename, delimiter, skip_header):
    delimiter = map_delimiter(delimiter)
    bytes_to_read = 1024 * 10  # for analysis
    rows = []
    with open(filename) as csvfile:
        if delimiter is None:
            sniffer = csv.Sniffer()
            content_to_analyze = csvfile.read(bytes_to_read)
            dialect = sniffer.sniff(content_to_analyze)
            csvfile.seek(0)
            csv_reader = csv.reader(csvfile, dialect=dialect)
        else:
            csv_reader = csv.reader(csvfile, delimiter=delimiter)

        for row in csv_reader:
            if not row:
                continue
            rows.append(row)
    if skip_header:
        rows = rows[1:]
    return rows


def parse_csv(rows, text_index, label_index):
    texts, labels = [], []
    for row in rows:
        if not row:
            continue
        if text_index and 0 <= text_index - 1 < len(row):
            texts.append(row[text_index - 1])
        if label_index and 0 <= label_index - 1 < len(row):
            labels.append(row[label_index - 1])
    return texts, labels


def validate_index(index):
    if not index:
        return False
    if isinstance(index, str) and not index.isdigit():
        return False
    index = int(index)
    if index < 1:
        raise Exception('Index should be greater or equal to 1')
    return True


def convert_to_int(value):
    return int(value)


def validate_and_convert_index(index):
    return convert_to_int(index) if validate_index(index) else None


def validate_and_convert_indices(text_index, label_index):
    text_index = validate_and_convert_index(text_index)
    label_index = validate_and_convert_index(label_index)
    if text_index is None:
        raise Exception('Text index is not set correctly')
    return text_index, label_index


def read_csv(filename, delimiter, skip_header, text_index, label_index):
    text_index, label_index = validate_and_convert_indices(text_index, label_index)
    rows = read_file(filename, delimiter, skip_header)
    if not rows:
        raise Exception('Cannot read the csv')
    texts, labels = parse_csv(rows, text_index, label_index)

    if not texts:
        raise Exception('Text could not be read from a file')
    return texts, labels



import unittest
from os import path

from ..base.common import cf_text_embeddings_package_path
from ..base.embeddings_model import AggregationMethod
from ..library import (cf_text_embeddings_glove, cf_text_embeddings_parse_csv,
                       cf_text_embeddings_tok_tok_tokenizer,
                       cf_text_embeddings_word2vec)


class TestLibrary(unittest.TestCase):
    def test_parse_csv(self):
        package_path = cf_text_embeddings_package_path()
        input_dict = {
            'input': path.join(package_path, 'corpus/basic.tsv'),
            'text_index': 2,
            'label_index': 3,
            'skip_header': 'true',
        }
        output_dict = cf_text_embeddings_parse_csv(input_dict)
        self.assertEqual(True, 'texts' in output_dict)
        self.assertEqual(True, 'labels' in output_dict)
        self.assertEqual(True, len(output_dict) == 2)
        self.assertEqual(True, len(output_dict['texts']) == 4)
        self.assertEqual(True, len(output_dict['labels']) == 4)

    def test_toktok_tokenizer(self):
        input_dict = {'texts': ['this is doc1'], 'to_lowercase': 'false'}
        output_dict = cf_text_embeddings_tok_tok_tokenizer(input_dict)
        self.assertEqual(True, 'texts' in output_dict)
        self.assertEqual(True, len(output_dict) == 1)
        self.assertEqual(True, len(output_dict['texts']) == 1)
        self.assertEqual(True, len(output_dict['texts'][0]) == 3)

    def test_word2vec_model(self):
        input_dict = {
            'texts': [['this', 'is', 'doc1'], ['this', 'is', 'doc2']],
            'labels': ['positive', 'negative'],
            'lang': 'test',
            'aggregation_method': AggregationMethod.summation.value,
        }
        output_dict = cf_text_embeddings_word2vec(input_dict)
        self.assertTrue('embeddings' in output_dict)
        self.assertTrue(len(output_dict['embeddings']) == 2)

    def test_glove_model(self):
        input_dict = {
            'texts': [['this', 'is', 'doc1'], ['this', 'is', 'doc2']],
            'labels': ['positive', 'negative'],
            'lang': 'test',
            'aggregation_method': AggregationMethod.summation.value,
        }
        output_dict = cf_text_embeddings_glove(input_dict)
        self.assertTrue('embeddings' in output_dict)
        self.assertTrue(len(output_dict['embeddings']) == 2)


if __name__ == '__main__':
    unittest.main()



from os import makedirs, path
from pathlib import Path

import numpy as np

PROJECT_DATA_DIR = path.join(str(Path.home()), '.cf_text_embeddings')


def map_checkbox_value(value):
    return value and value == 'true'


def load_numpy_array(path_):
    return np.load(path_)


def extract_map_invert_y(odt):
    unique_labels = odt.domain.class_var.values
    mapping = {v: k for v, k in enumerate(unique_labels)}
    y = np.array([mapping[int(label)] for label in odt.Y])
    return y, unique_labels


def to_int(num_string, default):
    try:
        num = int(num_string)
    except ValueError:
        return default
    return num


def to_float(num_string, default):
    try:
        num = float(num_string)
    except ValueError:
        return default
    return num


def ensure_dir(f):
    d = path.dirname(f)
    if not path.exists(d):
        makedirs(d, exist_ok=True)


def get_media_root():
    from mothra.settings import MEDIA_ROOT
    return MEDIA_ROOT


def save_numpy_array(dirname, filename, array):
    filepath = path.join(dirname, filename)
    ensure_dir(filepath)
    np.save(filepath, array)


def cf_text_embeddings_package_path():
    return path.dirname(path.dirname(path.abspath(__file__)))



import unittest
import warnings
from os import path

import numpy as np

from ..base.common import PROJECT_DATA_DIR
from ..base.embeddings_model import (TFIDF, AggregationMethod,
                                     EmbeddingsModelBert, EmbeddingsModelElmo,
                                     EmbeddingsModelLSI,
                                     EmbeddingsModelUniversalSentenceEncoder,
                                     EmbeddingsModelWord2Vec)


def word_tokens():
    return [[
        'Nintendo', 'has', 'announced', 'that', 'two', 'of', 'its', 'Switch', 'games', 'will', 'be',
        'updated.'
    ], ['Although', 'there’s', 'a', 'fierce', 'rivalry', 'between', 'smartphone', 'manufacturers.']]


def word_tokens_repeated_words():
    return [['Nintendo', 'has', 'that', 'and', 'of', 'fierce', 'Switch', 'games', 'and'],
            ['Although', 'there’s', 'and', 'fierce', 'rivalry', 'between', 'Switch']]


def sentence_tokens():
    return [['This is sentence1 of doc1', 'this is sentence2 of doc1'],
            ['This is sentence1 of doc2', 'this is sentence2 of doc2'],
            ['This is sentence1 of doc3']]


def doc_tokens():
    return [
        'Nintendo has announced that two of its Switch games will be updated.',
        'Although there’s a fierce rivalry between smartphone manufacturers.'
    ]


class EmbeddingsModelTest(unittest.TestCase):
    def test_word2vec_model(self):
        documents_tokens = word_tokens()
        embeddings_model = EmbeddingsModelWord2Vec('test')
        actual_X = embeddings_model.apply(documents_tokens, AggregationMethod.average.value, None)

        expected_X = np.array([[0.039, 0.026], [-0.045, -0.072]])
        self.assertEqual(True, np.allclose(expected_X, actual_X, atol=1e-03))

    def test_word2vec_model_not_exists(self):
        self.assertRaises(Exception, EmbeddingsModelWord2Vec, 'not_exists')

    def test_lsi_model_bow(self):
        np.random.seed(42)
        text = word_tokens()
        embeddings_model = EmbeddingsModelLSI(2, 1, train_on_tfidf=False, filter_extremes=False)
        actual_X = embeddings_model.apply(text, AggregationMethod.average.value, None)
        expected_X = np.array([[3.464], [2.828]])
        self.assertEqual(True, np.allclose(expected_X, actual_X, atol=1e-03))

    def test_lsi_model_tfidf(self):
        np.random.seed(42)
        text = word_tokens()
        embeddings_model = EmbeddingsModelLSI(2, 1, train_on_tfidf=True, filter_extremes=False)
        actual_X = embeddings_model.apply(text, AggregationMethod.average.value, None)
        expected_X = np.array([[1.], [1.]])
        self.assertEqual(True, np.allclose(expected_X, actual_X, atol=1e-03))

    def test_bert_model(self):
        if not path.exists(PROJECT_DATA_DIR):
            # This test is not run by Travis CI
            warnings.warn(
                'Test test_bert_model not executed, because you need to download model first')
            return

        documents_tokens = doc_tokens()
        embeddings_model = EmbeddingsModelBert('bert-base-uncased')
        embeddings = embeddings_model.apply(documents_tokens,
                                            aggregation_method=AggregationMethod.average.value)
        actual_X = embeddings[:, :2]

        expected_X = np.array([[-0.22724754, -0.42256614], [0.28956912, -0.16998145]])
        self.assertEqual(True, np.allclose(expected_X, actual_X, atol=1e-03))

    def test_elmo_model(self):
        english_elmo = path.join(PROJECT_DATA_DIR, 'models', 'en', 'elmo')
        if not path.exists(english_elmo):
            warnings.warn(
                'Test test_elmo_model not executed, because you need to download english elmo model'
            )
            return
        sentences = sentence_tokens()
        model = EmbeddingsModelElmo('en')
        embeddings = model.apply(sentences, aggregation_method=AggregationMethod.average.value)

        # The model returns different outputs when changing batch size
        # This happens due to states in LSTM
        actual_X = embeddings[:, :2]
        expected_X = np.array([[-0.13834494, -0.26584834]])
        self.assertEqual(True, np.allclose(expected_X, actual_X[0, :], atol=1e-03))
        self.assertEqual(3, actual_X.shape[0])

    def test_use_model(self):
        use_en = path.join(PROJECT_DATA_DIR, 'models', 'en', 'universal_sentence_encoder_english')
        if not path.exists(use_en):
            warnings.warn(
                'Test test_use_model not executed, because you need to download english use model')
            return
        sentences = sentence_tokens()
        model = EmbeddingsModelUniversalSentenceEncoder('en')
        embeddings = model.apply(sentences, aggregation_method=AggregationMethod.average.value)

        actual_X = embeddings[:, :2]
        expected_X = np.array([[0.06223562, -0.00187974], [0.06143629, -0.00408216],
                               [0.03041881, 0.04794494]])
        self.assertEqual(True, np.allclose(expected_X, actual_X, atol=1e-03))
        self.assertEqual(3, actual_X.shape[0])


class TFIDFTest(unittest.TestCase):
    def test_word2vec_model_with_tfidf(self):
        documents_tokens = word_tokens()
        tfidf_model = TFIDF()
        tfidf_model.train(documents_tokens)

        embeddings_model = EmbeddingsModelWord2Vec('test')
        actual_X = embeddings_model.apply(documents_tokens, AggregationMethod.average.value,
                                          tfidf_model)

        expected_X = np.array([[0.01139571, 0.00762929], [-0.0161487, -0.02562849]])
        self.assertEqual(True, np.allclose(expected_X, actual_X, atol=1e-03))

    def test_tfidf_model_calculate_document_weights(self):
        documents_tokens = word_tokens_repeated_words()
        tfidf_model = TFIDF()
        tfidf_model.train(documents_tokens)
        documents_weights_expected = {
            'Nintendo': 0.4472135954999579,
            'games': 0.4472135954999579,
            'has': 0.4472135954999579,
            'of': 0.4472135954999579,
            'that': 0.4472135954999579
        }
        documents_weights_actual = tfidf_model.calculate_document_weights(documents_tokens[0])
        for token_name in documents_weights_expected:
            self.assertEqual(documents_weights_expected[token_name],
                             documents_weights_actual[token_name])
        self.assertEqual(len(documents_weights_expected), len(documents_weights_actual))


if __name__ == '__main__':
    unittest.main()



import argparse
import glob
import zipfile
from os import path, remove, rename

import gitlab
import transformers

import requests

from .base.common import PROJECT_DATA_DIR, ensure_dir

GITLAB_URL = 'https://repo.ijs.si/'
GITLAB_PROJECT = 'vpodpecan/cf_text_embeddings_models'
REF = 'master'


def download_model(gitlab_path, local_filename):
    url = path.join(GITLAB_URL, GITLAB_PROJECT, 'raw', REF, gitlab_path)
    filepath, filename = path.split(local_filename)
    temp_filename = path.join(filepath, filename + '_temp')
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(temp_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
        rename(temp_filename, local_filename)  # handle partialy downloaded files


def download_bert_model():
    print('Downloading Bert Model')
    model_class = transformers.BertModel
    pretrained_weights_list = [
        'bert-base-uncased',
        'bert-base-multilingual-uncased',
        'distilbert-base-multilingual-cased',
    ]
    for pretrained_weights in pretrained_weights_list:
        print('Downloading', pretrained_weights)
        model_class.from_pretrained(pretrained_weights)


def unzip_file(zip_filepath):
    directory_to_extract_to = path.dirname(zip_filepath)
    with zipfile.ZipFile(zip_filepath, 'r') as zip_ref:
        zip_ref.extractall(directory_to_extract_to)


def unzip_files(file_paths):
    for file_path in file_paths:
        print('Unzipping %s' % (file_path))
        unzip_file(file_path)


def remove_files(file_paths):
    for file_path in file_paths:
        print('Removing zip %s' % file_path)
        remove(file_path)


def download_models_for_language(project, models_dir, language):
    print('Downloading models for %s' % language)

    models_tree = project.repository_tree(path='models/%s' % language, ref=REF)
    n_models = len(models_tree)

    language_dir = path.join(models_dir, 'models', language)
    zipped_models = [f for f in glob.glob(path.join(language_dir, "*.zip"))]
    downloaded_models = [path.basename(model_path) for model_path in zipped_models]

    for i, model_tree in enumerate(models_tree, 1):
        model_name = model_tree['name']
        gitlab_path = model_tree['path']
        file_path = path.join(models_dir, gitlab_path)
        ensure_dir(file_path)
        if model_name not in downloaded_models:
            print('%d/%d Downloading model %s to %s' % (i, n_models, gitlab_path, file_path))
            download_model(gitlab_path, file_path)
            zipped_models.append(file_path)
    return zipped_models


def process_models_for_language(project, models_dir, language):
    file_paths = download_models_for_language(project, models_dir, language)
    unzip_files(file_paths)
    remove_files(file_paths)


def fetch_languages(project):
    languages_tree = project.repository_tree(path='models', ref=REF)
    languages = sorted([language_tree['name'] for language_tree in languages_tree])
    languages += ['all']
    return languages


def list_languages(languages):
    print('Available languages:')
    for i, language in enumerate(languages, 1):
        print('%d. %s' % (i, language))


def main():
    # anonymous gitlab instance, read-only for public resources
    gl = gitlab.Gitlab(GITLAB_URL)
    project = gl.projects.get(GITLAB_PROJECT)
    models_dir = PROJECT_DATA_DIR
    ensure_dir(models_dir)

    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--list", action="store_true", help="list available languages")
    parser.add_argument("-d", "--download", type=str,
                        help="download models for certain language code")
    args = parser.parse_args()

    languages = fetch_languages(project)

    if args.list:
        list_languages(languages)
    elif args.download in languages:
        language = args.download
        if language == 'all':
            for language in languages:
                process_models_for_language(project, models_dir, language)
        else:
            process_models_for_language(project, models_dir, language)
        if language == 'multi':
            download_bert_model()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import shutil
import tempfile
import uuid
from os import path

from django.shortcuts import render

from .base.common import (ensure_dir, extract_map_invert_y, get_media_root,
                          save_numpy_array)


def cf_text_embeddings_export_dataset(request, input_dict, output_dict, widget):
    dataset = input_dict.get('dataset')
    if dataset is None:
        raise Exception('There is no dataset in the input')

    archive_filename = str(uuid.uuid4())
    archive_download_filepath = path.join(str(request.user.id), archive_filename + '.zip')
    archive_local_filepath = path.join(get_media_root(), str(request.user.id), archive_filename)
    ensure_dir(archive_local_filepath)

    y, _ = extract_map_invert_y(dataset)
    td = tempfile.TemporaryDirectory(dir=path.join(get_media_root(), str(request.user.id)))
    save_numpy_array(td.name, 'x.npy', dataset.X)
    save_numpy_array(td.name, 'y.npy', y)

    shutil.make_archive(archive_local_filepath, 'zip', td.name)
    td.cleanup()

    output_dict['filename'] = archive_download_filepath
    return render(request, 'visualizations/string_to_file.html', {
        'widget': widget,
        'input_dict': input_dict,
        'output_dict': output_dict
    })



import logging
from enum import Enum
from os import path

import tensorflow.compat.v1 as tf
import tensorflow_hub as hub
import tf_sentencepiece  # NOQA # pylint: disable=unused-import
from elmoformanylangs import Embedder
from gensim import corpora
from gensim.corpora import Dictionary
from gensim.models import LsiModel, TfidfModel
from gensim.models.doc2vec import Doc2Vec
from gensim.models.keyedvectors import (FastTextKeyedVectors,
                                        Word2VecKeyedVectors)
from transformers import BertModel, pipeline

import numpy as np

import numpy as np

from .common import PROJECT_DATA_DIR, cf_text_embeddings_package_path

# disable logs because they are output as messages in clowdflows
logging.getLogger('gensim').setLevel(logging.ERROR)
logging.getLogger('elmoformanylangs').setLevel(logging.ERROR)
logging.getLogger('transformers').setLevel(logging.ERROR)
tf.get_logger().setLevel('ERROR')


class AggregationMethod(Enum):
    average = 'average'
    summation = 'summation'


def aggregate_document_embeddings(document_embedding, aggregation_method):
    if aggregation_method == AggregationMethod.summation.value:
        return np.sum(document_embedding, axis=0)
    if aggregation_method == AggregationMethod.average.value:
        return np.average(document_embedding, axis=0)
    raise Exception('%s aggregation method is not supported' % aggregation_method)


def cf_text_embeddings_model_path(lang, model_name):
    models_dir = cf_text_embeddings_package_path() if lang == 'test' else PROJECT_DATA_DIR
    lang = '' if lang is None else lang
    model_name = '' if model_name is None else model_name
    return path.join(models_dir, 'models', lang, model_name)


class EmbeddingsModelBase:
    def __init__(self, lang):
        self._lang = lang
        self._model_name = self.extract_model_name(lang)
        if self._model_name is None:
            raise Exception('Model for %s language is not supported' % lang)
        self._path = cf_text_embeddings_model_path(self._lang, self._model_name)
        self._model = None
        self._tfidf_dict = None
        self._tfidf_model = None

    def _load_model(self):
        return Word2VecKeyedVectors.load(self._path, mmap='r')

    @staticmethod
    def supported_models():
        return {}

    @classmethod
    def extract_model_name(cls, lang):
        supported_models = cls.supported_models()
        return supported_models.get(lang)

    @staticmethod
    def _tokens_to_embeddings(model, documents_tokens, aggregation_method, tfidf):
        embeddings = np.zeros((len(documents_tokens), model.vector_size))
        for i, document_tokens in enumerate(documents_tokens):
            documents_weights = {}
            if tfidf is not None:
                documents_weights = tfidf.calculate_document_weights(document_tokens)

            n_tokens = 0
            for token in document_tokens:
                if token not in model.vocab:
                    continue
                if aggregation_method in {
                        AggregationMethod.summation.value, AggregationMethod.average.value
                }:
                    # token vector is multiplied by TFIDF weigth if available
                    embeddings[i] += model[token] * documents_weights.get(token, 1)
                    n_tokens += 1
            if aggregation_method == AggregationMethod.average.value and n_tokens > 0:
                embeddings[i] /= n_tokens
        return embeddings

    def apply(self, texts, aggregation_method=AggregationMethod.average.value, tfidf=None):
        try:
            self._model = self._load_model()
        except FileNotFoundError:
            raise FileNotFoundError('Cannot find the model. Download it, if available.')
        embeddings = self._tokens_to_embeddings(self._model, texts,
                                                aggregation_method=aggregation_method, tfidf=tfidf)
        return embeddings

    def __getstate__(self):
        state = self.__dict__
        state.pop('_model')
        return state

    def __setstate__(self, state):
        # pylint: disable=W0201
        state['_model'] = None
        self.__dict__ = state


class EmbeddingsModelWord2Vec(EmbeddingsModelBase):
    @staticmethod
    def supported_models():
        return {
            # https://drive.google.com/file/d/0B7XkCwpI5KDYNlNUTTlSS21pQmM
            'en': 'GoogleNews-vectors-negative300.wv.bin',
            # https://github.com/uchile-nlp/spanish-word-embeddings
            'es': 'SBW-vectors-300-min5.wv.bin',
            # http://vectors.nlpl.eu/repository/
            'sl': 'word2vec_si.wv',
            # http://vectors.nlpl.eu/repository/
            'hr': 'word2vec_hr.wv',
            # http://vectors.nlpl.eu/repository/
            'de': 'word2vec_de.wv',
            # http://vectors.nlpl.eu/repository/
            'ru': 'word2vec_ru.wv',
            # http://vectors.nlpl.eu/repository/
            'lv': 'word2vec_lv.wv',
            # http://vectors.nlpl.eu/repository/
            'ee': 'word2vec_ee.wv',
            'test': 'word2vec_test_model.bin'
        }


class EmbeddingsModelGloVe(EmbeddingsModelBase):
    @staticmethod
    def supported_models():
        return {
            # https://nlp.stanford.edu/projects/glove/
            'en': 'glove.6B.300d.wv.bin',
            # https://github.com/dccuchile/spanish-word-embeddings
            'es': 'glove_es.wv',
            # https://deepset.ai/german-word-embeddings
            'de': 'glove_de.wv',
            # internally glove models are handeled the same as word2vec
            'test': 'word2vec_test_model.bin'
        }


class EmbeddingsModelFastText(EmbeddingsModelBase):
    @staticmethod
    def supported_models():
        return {
            # https://github.com/facebookresearch/MUSE
            'en': 'wiki.multi.en.wv',
            # https://github.com/facebookresearch/MUSE
            'es': 'wiki.multi.es.wv',
            # https://github.com/facebookresearch/MUSE
            'de': 'wiki.multi.de.wv',
            # https://github.com/facebookresearch/MUSE
            'ru': 'wiki.multi.ru.wv',
            # https://fasttext.cc/docs/en/crawl-vectors.html
            'lv': 'fasttext_lv.wv',
            # https://fasttext.cc/docs/en/crawl-vectors.html
            'lt': 'fasttext_lt.wv',
            # https://github.com/facebookresearch/MUSE
            'ee': 'wiki.multi.ee.wv',
            # https://github.com/facebookresearch/MUSE
            'sl': 'wiki.multi.sl.wv',
            # https://github.com/facebookresearch/MUSE
            'hr': 'wiki.multi.hr.wv',
        }

    def _load_model(self):
        return FastTextKeyedVectors.load(self._path, mmap='r')


class EmbeddingsModelFastTextHr(EmbeddingsModelFastText):
    @staticmethod
    def supported_models():
        # https://www.clarin.si/repository/xmlui/handle/11356/1205
        return {
            'hr': 'embed.hr-token.ft.sg.vec.wv',
        }


class EmbeddingsModelFastTextSl(EmbeddingsModelFastText):
    @staticmethod
    def supported_models():
        # https://www.clarin.si/repository/xmlui/handle/11356/1204
        return {
            'sl': 'embed.sl-token.ft.sg.wv',
        }


class EmbeddingsModelFastTextEmbedia(EmbeddingsModelFastText):
    @staticmethod
    def supported_models():
        # https://kt-cloud.ijs.si/index.php/apps/files/?dir=/embeddia/Data/embeddings_models/fastText&fileid=156407
        return {
            'sl': 'sl-embs300.vec.wv',
        }


class EmbeddingsModelDoc2Vec(EmbeddingsModelBase):
    @staticmethod
    def supported_models():
        return {
            # https://github.com/jhlau/doc2vec
            'en': 'doc2vec.bin',
        }

    def _load_model(self):
        return Doc2Vec.load(self._path, mmap='r')

    def _tokens_to_embeddings(self, model, documents_tokens, aggregation_method, tfidf):
        document_embeddings = np.zeros((len(documents_tokens), model.vector_size))
        for i, document_tokens in enumerate(documents_tokens):
            if not document_tokens:
                continue
            document_embeddings[i] = model.infer_vector(document_tokens)
        return document_embeddings


class EmbeddingsModelUniversalSentenceEncoder(EmbeddingsModelBase):
    def __init__(self, lang):
        super().__init__(lang)
        self.embeddings_size = 512

    @staticmethod
    def supported_models():
        return {
            # https://tfhub.dev/google/universal-sentence-encoder/2
            'en': 'universal_sentence_encoder_english',
            # https://tfhub.dev/google/universal-sentence-encoder-xling/en-de/1
            'de': 'universal_sentence_encoder_german',
            # https://tfhub.dev/google/universal-sentence-encoder-xling/en-es/1
            'es': 'universal_sentence_encoder_spanish',
        }

    def apply(self, texts, aggregation_method=AggregationMethod.average.value, tfidf=None):
        # To make tf 2.0 compatible with tf1.0 code, we disable the tf2.0 functionalities
        tf.disable_eager_execution()
        path_ = cf_text_embeddings_model_path(self._lang, self._model_name)

        # Create graph and finalize (finalizing optional but recommended).
        g = tf.Graph()
        with g.as_default():
            # We will be feeding 1D tensors of text into the graph.
            text_input = tf.placeholder(dtype=tf.string, shape=[None])
            model = hub.Module(path_)
            embedded_text = model(text_input)
            init_op = tf.group([tf.global_variables_initializer(), tf.tables_initializer()])
        g.finalize()

        # Create session and initialize.
        session = tf.Session(graph=g)
        session.run(init_op)

        embeddings = np.zeros((len(texts), self.embeddings_size))
        for i, document_tokens in enumerate(texts):
            if not document_tokens:
                continue
            document_embedding = session.run(embedded_text, feed_dict={text_input: document_tokens})
            embeddings[i] = aggregate_document_embeddings(document_embedding, aggregation_method)
        return embeddings


class EmbeddingsModelElmo(EmbeddingsModelBase):
    def __init__(self, lang):
        super().__init__(lang)
        self.embeddings_size = 1024

    @staticmethod
    def supported_models():
        return {
            # http://vectors.nlpl.eu/repository/#
            'en': 'elmo',
            # http://vectors.nlpl.eu/repository/#
            'sl': 'elmo',
            # http://vectors.nlpl.eu/repository/#
            'es': 'elmo',
            # http://vectors.nlpl.eu/repository/#
            'ru': 'elmo',
            # http://vectors.nlpl.eu/repository/#
            'hr': 'elmo',
            # http://vectors.nlpl.eu/repository/#
            'ee': 'elmo',
            # http://vectors.nlpl.eu/repository/#
            'lv': 'elmo',
            # http://vectors.nlpl.eu/repository/#
            'de': 'elmo',
        }

    def _load_model(self):
        return Embedder(self._path)

    def _tokens_to_embeddings(self, model, documents_tokens, aggregation_method, tfidf):
        model.model.eval()
        embeddings = np.zeros((len(documents_tokens), self.embeddings_size))
        batch_size = 100

        doc_index = 0
        for i in range(0, len(documents_tokens), batch_size):
            documents_embeddings = model.sents2elmo(documents_tokens[i:i + batch_size])
            for document_embeddings in documents_embeddings:
                embeddings[doc_index] = aggregate_document_embeddings(document_embeddings,
                                                                      aggregation_method)
                doc_index += 1
        return embeddings


class EmbeddingsModelHuggingface(EmbeddingsModelBase):
    def __init__(self, model_name, tokenizer, vector_size):
        self.model_name = model_name
        self.tokenizer = tokenizer
        self.vector_size = vector_size
        self._model = None

    def _load_model(self):
        self._model = pipeline('feature-extraction', model=self.model_name,
                               tokenizer=self.tokenizer)
        return self._model

    def _tokens_to_embeddings(self, model, documents_tokens, aggregation_method, tfidf):
        embeddings = np.zeros((len(documents_tokens), self.vector_size))
        for i, document_tokens in enumerate(documents_tokens):
            document_embedding = np.array(self._model(document_tokens)[0])
            if document_embedding.size > 0:
                embeddings[i] = aggregate_document_embeddings(document_embedding,
                                                              aggregation_method)
        return embeddings


class EmbeddingsModelBert(EmbeddingsModelHuggingface):
    def __init__(self, model):
        supported_models = self.supported_models()
        model_conf = supported_models[model]
        self.model_name = model_conf['model_name']
        self.tokenizer = model_conf['tokenizer']
        self.vector_size = model_conf['vector_size']
        super().__init__(self.model_name, self.tokenizer, self.vector_size)

    @staticmethod
    def supported_models():
        return {
            # model source: https://huggingface.co/transformers/pretrained_models.html
            'bert-base-uncased': {
                'model_name': 'bert-base-uncased',
                'tokenizer': 'bert-base-uncased',
                'vector_size': 768,
            },
            'bert-base-multilingual-uncased': {
                'model_name': 'bert-base-multilingual-uncased',
                'tokenizer': 'bert-base-multilingual-uncased',
                'vector_size': 768,
            },
            'distilbert-base-multilingual-cased': {
                'model_name': 'distilbert-base-multilingual-cased',
                'tokenizer': 'distilbert-base-multilingual-cased',
                'vector_size': 768,
            },
        }


class EmbeddingsModelBertEmbeddia(EmbeddingsModelBert):
    def __init__(self, model):
        super().__init__(model=model)

    def _load_model(self):
        model_path = cf_text_embeddings_model_path('multi', self.model_name)
        model = BertModel.from_pretrained(model_path)
        self._model = pipeline('feature-extraction', model=model, tokenizer=self.tokenizer)
        return self._model

    @staticmethod
    def supported_models():
        return {
            # https://www.clarin.si/repository/xmlui/handle/11356/1317
            'CroSloEngualBert': {
                'model_name': 'CroSloEngualBert',
                'tokenizer': 'bert-base-uncased',
                'vector_size': 768,
            }
        }


class EmbeddingsModelLSI(EmbeddingsModelBase):
    def __init__(self, num_topics, decay, train_on_tfidf, filter_extremes):
        self.num_topics = num_topics
        self.decay = decay
        self.train_on_tfidf = train_on_tfidf
        self.filter_extremes = filter_extremes

    def apply(self, texts, aggregation_method=AggregationMethod.average.value, tfidf=None):
        dct = corpora.Dictionary(texts)
        if self.filter_extremes:
            dct.filter_extremes()
            if len(dct) == 0:
                raise Exception('Disable filter extremes as it removes all tokens')
        corpus = [dct.doc2bow(line) for line in texts]
        if self.train_on_tfidf:
            tfidf_model = TfidfModel(corpus)
            corpus = tfidf_model[corpus]
        self._model = LsiModel(corpus=corpus, id2word=dct, num_topics=self.num_topics,
                               decay=self.decay)
        embeddings = np.array([[el[1] for el in self._model[document]] for document in corpus])
        return embeddings


class TFIDF:
    def __init__(self):
        self.model = None
        self.dictionary = None
        self.id2token = {}

    def train(self, documents_tokens):
        self.dictionary = Dictionary(documents_tokens)  # fit dictionary
        # id2token is lazy initialized in gensim
        self.id2token = {v: k for k, v in self.dictionary.token2id.items()}
        corpus = [self.dictionary.doc2bow(document_tokens) for document_tokens in documents_tokens]
        self.model = TfidfModel(corpus)
        return self.model

    def calculate_document_weights(self, document_tokens):
        token_weights = {}
        document_bow = self.dictionary.doc2bow(document_tokens)
        for token_id, weigth in self.model[document_bow]:
            token_name = self.id2token.get(token_id, '')
            token_weights[token_name] = weigth
        return token_weights



import unittest

from ..base.tokenizers import (punkt_tokenizer, token_filtering,
                               toktok_tokenizer)


class TestTokenizers(unittest.TestCase):
    def test_toktok_tokenizer(self):
        texts = ['this is some text', 'this is some other text']
        texts_tokenized = toktok_tokenizer(texts)
        for i, text in enumerate(texts):
            tokens = text.split(' ')
            for j, token in enumerate(tokens):
                self.assertEqual(token, texts_tokenized[i][j])

    def test_toktok_tokenizer_empty_input(self):
        texts = ['']
        texts_tokenized = toktok_tokenizer(texts)
        self.assertEqual(1, len(texts_tokenized))
        self.assertEqual(0, len(texts_tokenized[0]))

    def test_punkt_tokenizer(self):
        texts = ['this is some text', 'this is some other text']
        texts_tokenized = punkt_tokenizer(texts, language='english')
        for i, text in enumerate(texts):
            self.assertEqual([text], texts_tokenized[i])

    def test_punkt_tokenizer_empty_input(self):
        texts = ['']
        texts_tokenized = punkt_tokenizer(texts, language='english')
        self.assertEqual(1, len(texts_tokenized))
        self.assertEqual(0, len(texts_tokenized[0]))


class TestTokenFiltering(unittest.TestCase):
    def test_token_filtering(self):
        documents_tokens = [['this', 'is', 'some', 'text'], ['this', 'is', 'other', 'text']]
        documents_tokens_filtered = token_filtering(documents_tokens, filter_tokens=None)

        for i, _ in enumerate(documents_tokens):
            for j, _ in enumerate(documents_tokens[i]):
                self.assertEqual(documents_tokens[i][j], documents_tokens_filtered[i][j])

    def test_token_filtering_with_tokens_to_remove(self):
        documents_tokens = [['this', 'is', ',', 'some', 'text', '.']]
        documents_tokens_expected = [['this', 'is', 'some', 'text']]
        documents_tokens_actual = token_filtering(documents_tokens, filter_tokens=None)

        for i, _ in enumerate(documents_tokens_expected):
            for j, _ in enumerate(documents_tokens_expected[i]):
                self.assertEqual(documents_tokens_expected[i][j], documents_tokens_actual[i][j])

    def test_token_filtering_filter_token_empyty_string(self):
        documents_tokens = [['this', 'is', ',', 'some', 'text', '.']]
        documents_tokens_expected = [['this', 'is', 'some', 'text']]
        documents_tokens_actual = token_filtering(documents_tokens, filter_tokens='')

        for i, _ in enumerate(documents_tokens_expected):
            for j, _ in enumerate(documents_tokens_expected[i]):
                self.assertEqual(documents_tokens_expected[i][j], documents_tokens_actual[i][j])

    def test_token_filtering_with_custom_filter(self):
        documents_tokens = [['<SOS>', 'this', 'is', 'some', 'text', '<EOS>']]
        documents_tokens_expected = [['this', 'is', 'some', 'text']]
        documents_tokens_actual = token_filtering(documents_tokens, filter_tokens='<SOS>\n<EOS>')

        for i, _ in enumerate(documents_tokens_expected):
            for j, _ in enumerate(documents_tokens_expected[i]):
                self.assertEqual(documents_tokens_expected[i][j], documents_tokens_actual[i][j])

    def test_token_filtering_all_tokens_filtered(self):
        documents_tokens = [['.', ',', ')']]
        documents_tokens_expected = [[]]
        documents_tokens_actual = token_filtering(documents_tokens, filter_tokens=None)

        for i, _ in enumerate(documents_tokens_expected):
            for j, _ in enumerate(documents_tokens_expected[i]):
                self.assertEqual(documents_tokens_expected[i][j], documents_tokens_actual[i][j])


if __name__ == '__main__':
    unittest.main()



import os
import re

from setuptools import setup

with open('cf_text_embeddings/version.py', "rt") as f:
    version_content = f.read()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, version_content, re.M)
    if mo:
        version_string = mo.group(1)
    else:
        raise RuntimeError("Unable to find version string")

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()
    # remove multiple spaces because of pypi requirement
    README = re.sub(' +', '', README)

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='cf_text_embeddings',
    version=version_string,
    packages=['cf_text_embeddings'],
    include_package_data=True,
    license='MIT License',
    description='Text Embeddings for ClowdFlows',
    long_description=README,
    url='https://github.com/xflows/cf_text_embeddings',
    author='Roman Orac',
    author_email='orac.roman@gmail.com',
    install_requires=requirements,
)


def download_nltk_requirements():
    try:
        import nltk  # NLTK should be imported after it is installed
        nltk.download('punkt')
    except Exception:
        print(('Warning: NLTK punkt languages weren\'t downloaded,'
               'run: python -m nltk.downloader punkt'))


def post_install():
    download_nltk_requirements()


post_install()
